#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <netdb.h> 
#include <sock.h>
#include <setup.h>

#ifdef  HAVE_SYS_TIMES_H
# include <sys/times.h>
#endif/*HAVE_SYS_TIMES_H*/
 
#if  TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif/*TIME_WITH_SYS_TIME*/ 

#define RESBUF  9000

/* local prototypes */
int mknblock( int sock, int nonblock ); 

int
WACKYsocket( char *hostname, char *portname, socklen_t *addrlen )
{
  int sock, port, err;
  int reuse = 1;
  struct hostent hent;
  struct hostent *hp;
  int    herrno;
  struct sockaddr_in serv;
  struct in_addr     addr;
  struct servent     *sp;
  char   *protocol = "tcp";
  struct addrinfo hints, *info, *save;
 
  bzero(&hints, sizeof(struct addrinfo));
  hints.ai_flags = AI_PASSIVE;
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
 
  if(( err = getaddrinfo( hostname, portname, &hints, &info )) != 0 ){
    joe_fatal("unable to acquire address info");
  }
  save = info;

  if(( sock = socket( info->ai_family, info->ai_socktype, info->ai_protocol )) < 0 ){
    switch( errno ){
      case EPROTONOSUPPORT: { joe_error("unsupported protocol");      break; }
      case EMFILE:          { joe_error("descriptor table is full");  break; }
      case ENFILE:          { joe_error("system file table is full"); break; }
      case EACCES:          { joe_error("permission denied");         break; }
      case ENOBUFS:         { joe_error("insufficient buffer space"); break; }
      default:              { joe_error("unknown socket error");      break; }
    } return -1;   
  }

  if( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse, sizeof( reuse )) < 0 ){
    switch( errno ){
      case EBADF:           { joe_error("invalid descriptor");       break; } 
      case ENOTSOCK:        { joe_error("not a socket");             break; }
      case ENOPROTOOPT:     { joe_error("not a protocol option");    break; }
      case EFAULT:          { joe_error("setsockopt unknown error"); break; }
      default:              { joe_error("unknown sockopt error");    break; }
    }
  }
  
  memset( &serv, 0, sizeof( struct sockaddr_in ));
  serv.sin_family = AF_INET;

  if( !strncasecmp( hostname, "*", 1 )){
    serv.sin_addr.s_addr = htonl( INADDR_ANY );   
  }
  else{
    char *buf = (char*)xmalloc( RESBUF );
    if(( gethostbyname_r( hostname,
                        (struct hostent *)buf,
                        buf + sizeof(struct hostent),
                        RESBUF - sizeof(struct hostent),
                        &hp,
                        &h_errno )) < 0 )
    hp = NULL; 
  }

  if( hp != NULL ){
    serv.sin_addr.s_addr = ((struct in_addr*)hp->h_addr_list[0])->s_addr; 
  }

  if( serv.sin_addr.s_addr == INADDR_NONE ){
    joe_fatal("WACKYsocket: flawed address");
  } 

  if(( port = atoi( portname )) == 0 ){
    if(( sp = getservbyname( portname, protocol )) == NULL )
      joe_error( "WACKYsocket %s", portname );
    serv.sin_port = sp->s_port;
  } 
  else{
    serv.sin_port = htons( atoi( portname ));  
  }

  if(( err = bind( sock, (struct sockaddr*)&serv, sizeof( serv )) < 0 )){
    switch(errno) {
      case EBADF:  { joe_error("invalid descriptor");          break; }
      case EINVAL: { joe_error("already bound to an address"); break; }
      case EACCES: { joe_error("protected address");           break; }
      default:     { joe_error("unknown bind error");          break; }
    } return -1;
  }

  if(( err = listen( sock, 10 )) < 0 ){
    switch( errno ){
      case EBADF:      { joe_error("invalid descriptor");      break; }
      case ENOTSOCK:   { joe_error("descriptor not a socket"); break; }
      case EOPNOTSUPP: { joe_error("unsupported operation");   break; }
      default:         { joe_error("unknown listen error");    break; }
    }  
  } 

  if( addrlen ){ *addrlen = info->ai_addrlen; }
  freeaddrinfo( save ); 
 
  return sock;
}

/**
 * local function;
 * checks the  socket for  inactivity,
 * returns -1 on failure, 0 on success
 */
int
WACKYsocket_check( CONN *C )
{
  int res;
  struct timeval timeout;
 
  FD_ZERO( &C->rset );
  FD_SET( C->sock, &C->rset );
  memset((void *)&timeout, '\0', sizeof( struct timeval ));
 
  if(( mknblock( C->sock, TRUE )) < 0 ){
    joe_error( "WACKYsocket: unable to set socket to non-blocking." );
    return -1;
  }
 
  timeout.tv_sec  = (my.timeout > 0)?my.timeout:15;
  timeout.tv_usec = 0;
 
  if(( res = select( C->sock+1, &C->rset, NULL, NULL, &timeout )) < 1 ){
    return -1;
  }
 
  if(( mknblock( C->sock, FALSE )) < 0 ){
    joe_error( "WACKYsocket: unable to set socket to non-blocking." );
    return -1;
  }
  FD_CLR( C->sock, &C->rset );
 
  return 0;
} 

/**
 * local function;
 * set socket to non-blocking to
 * allow select on it.
 */
int
mknblock( int sock, int nonblock )
{
#if HAVE_FCNTL_H
  int flags;
  int retval;
 
  flags = fcntl( sock, F_GETFL, 0 );
  if( flags < 0 ){
    joe_error("fcntl");
    return -1;
  }
  if( nonblock ){
    flags |=  O_NDELAY;
  }
  else{
    flags &= ~O_NDELAY;
  }
  retval = fcntl( sock, F_SETFL, flags);
  if( retval < 0 ){
    perror("fcntl");
    return -1;
  }
#elif defined( FIONBIO )
  ioctl_t status;
 
  status = nb ? 1 : 0;
  return ioctl( sd, FIONBIO, &status );
#endif
} 

int
WACKYaccept( int sock, struct sockaddr_in *clnt, socklen_t *len )
{
  int i;
 
retry:
  if(( i = accept( sock, (struct sockaddr*)&clnt, len )) < 0) {
#ifdef  EPROTO
  if( errno == EPROTO || errno == ECONNABORTED )
#else
  if( errno == ECONNABORTED )
#endif
    goto retry;
  else
    joe_error("accept error");
  }
  return i;
} 

ssize_t
WACKYreadline( CONN *C, char *ptr, size_t len )
{
  int n;
  
  do{
    n = read( C->sock, ptr, 1 ); 
  } while( n > 0 && *ptr++ != '\n' );
  
  *ptr++=0;
  return( n > 0 ); 
}

/**
 * returns void
 * frees ssl resources if using ssl and
 * closes the connection and the socket.
 */
void
WACKYclose( CONN *C )
{
  close( C->sock );
  return;
} 
